# Python3 assembler for Tiger ISA
# usage: asm <Desktop/TigerSBC-relative_path/yourprog.asm> e.g. apps/yourapp.asm
import sys # for command line arguments
import math # facilitate generation of lookup tables

srcfile = sys.argv[1] # source filename
_includes_ = [] # for debug in case of an error report

# 12-bit fundamentals
ALLBITS = 0xfff
SIGNBIT = 0x800
MAXPOS  = 0x7ff
ram = [0 for i in range(0x1000)] # zeroed RAM image

# return a 12-bit evaluation of expr
def bits12(expr):
    assert expr <= ALLBITS
    assert expr >= -SIGNBIT
    return expr & ALLBITS # suppress sign-extension

# slot 1 opcode images
NOP  = 0x000
LIT  = 0x100
LD   = 0x200
ST   = 0x300
TT   = 0x400
INV  = 0x500
INC  = 0x600
ASR  = 0x700
ADX  = 0x800
ADD  = 0x900
BAND = 0xa00
BXOR = 0xb00
JZ   = 0xc00
JMP  = 0xd00
CALL = 0xe00
RET  = 0xf00

# global state variables
slots  = 0 # number of instructions in current packet of 3
packet = 0 # uncommited image, initial NOPs
pptr   = 0 # packet pointer
lptr   = 1 # literal pointer
passn  = 0 # 0: collect forward references, 1: generate final code

# flush packet to ram
def flush_packet():
    global slots, packet, pptr, lptr
    ram[pptr] = packet
    pptr = lptr
    lptr += 1
    packet = 0
    slots = 0

# align to next packet
def align():
    if slots:
        flush_packet()

# enforce uniqueness of a new symbol
def assert_unique(string):
        exists = string in globals()
        if exists:
            print(string + ' already exists!')
        assert not exists

# define a unique memory label in pass 0
def lbl(string):
    align()
    if passn == 0:
        assert_unique(string)
        exec('global ' + string + '; ' + string + ' = pptr')

# define a unique 12-bit numeric symbol in pass 0
def def12(string, expr):
    if passn == 0:
        assert_unique(string)
        exec('global ' + string + '; ' + string + ' = ' + str(bits12(expr)))

# define a unique general numeric symbol in pass 0
def defnum(string, expr):
    if passn == 0:
        assert_unique(string)
        exec('global ' + string + '; ' + string + ' = ' + str(expr))

# forgive a forward reference in pass 0
def fwd(string):
    if passn == 0:
        return 0
    else:
        return eval(string)

# assemble opcode
def asmop(opcode):
    global slots, packet
    packet |= opcode >> (slots * 4)
    slots += 1
    if slots == 3:
        flush_packet()

def nop():
    asmop(NOP)

def lit(expr):
    global lptr
    ram[lptr] = bits12(expr)
    lptr += 1
    asmop(LIT)

def ld():
    asmop(LD)

def st():
    asmop(ST)

def tt():
    asmop(TT)

def inv():
    asmop(INV)

def inc():
    asmop(INC)

def asr():
    asmop(ASR)

def adx():
    asmop(ADX)

def add():
    asmop(ADD)

def band():
    asmop(BAND)

def bxor():
    asmop(BXOR)

def jz():
    asmop(JZ)

def jmp():
    asmop(JMP)
    align()

def call():
    asmop(CALL)
    align()

def ret():
    asmop(RET)
    align()

# lit macros

def lld(adr):
    lit(adr); ld()

def lst(adr):
    lit(adr); st()

def laddx(d):
    lit(d); addx()

def ladd(d):
    lit(d); add()

def lband(d):
    lit(d); band()

def lbxor(d):
    lit(d); bxor()

def ljz(adr):
    lit(adr); jz()

def ljmp(adr):
    lit(adr); jmp()

def lcall(adr):
    lit(adr); call()

# array access macros

def ald():
    add(); ld()

def lald(adr):
    lit(adr); ald()

def ast():
    add(); st()

def last(adr):
    lit(adr); ast()

# subtraction macros

def neg():
    inv(); inc()

def sub():
    neg(); add()

def lt():
    sub(); lband(SIGNBIT)

# shift left macro

def shl():
    tt(); add()

# assemble a list of data
def data(numlist):
    global pptr, lptr
    align()
    for i in numlist:
        ram[pptr] = bits12(i)
        pptr += 1
        lptr += 1

# assemble a zero-terminated ASCII string
def ascii(string):
    global pptr, lptr
    align()
    for i in range(len(string)):
        num = ord(string[i])
        assert num < 128
        ram[pptr] = num
        pptr += 1
        lptr += 1
    ram[pptr] = 0
    pptr += 1
    lptr += 1

# include a file
def include(string):
    align()
    _includes_.append(string) # log the source file as a clue to debugging any future errors
    with open(string, 'r') as f:
        for line in f:
            exec(line.lstrip()) # prevent unexpected indent errors
    align()

# assemble a file
def asm(string):
    global slots, packet, pptr, lptr, passn
    slots  = 0
    packet = 0
    pptr   = 0
    lptr   = 1
    passn  = 0
    include(string)
    slots  = 0
    packet = 0
    pptr   = 0
    lptr   = 1
    passn  = 1
    include(string)
    print(str(pptr) + ' 12-bit locations assembled')

# display memory from start_adr to end_adr-1
def dump_ram(start_adr, end_adr):
    while start_adr != end_adr:
        print((format(ram[start_adr], '03X') + ' '), end='')
        start_adr += 1
    print('')

# save entire memory for FPGA RAM initialization
def save_hex(string):
    with open(string, 'w') as f:
        for i in range(4096):
            f.write(format(ram[i], '03X') + '\n')

# save app_start..app_end-1 to output file
def save_app(string):
    tx_pair = bytearray(2) # little-endian data
    with open(string, 'wb') as f:
        # write the header
        tx_pair[0] = app_end & 0x3f
        tx_pair[1] = (app_end >> 6) & 0x3f
        f.write(tx_pair)
        # write the payload
        current_pair = app_start
        while current_pair != app_end:
            tx_pair[0] = ram[current_pair] & 0x3f
            tx_pair[1] = (ram[current_pair] >> 6) & 0x3f
            f.write(tx_pair)
            current_pair += 1

# timing characteristics
fclk       = 36000000 # Hz
tclk       = 1/fclk # seconds

# mmio addresses

# 48-bit timer, read-only
tl47       = 0xfff # r
tl35       = 0xffe # r
tl23       = 0xffd # r
tl11       = 0xffc # r, latches tl23, tl35, and tl47
t47        = 0xffb # r
t35        = 0xffa # r
t23        = 0xff9 # r
t11        = 0xff8 # r

# write-only dest for stack drops
xxxxxxxxxx = tl47  # w

def drop():
    lst(xxxxxxxxxx)

# 28-bit macc
norm32     = 0xff7 # r
norm16     = 0xff6 # r
norm8      = 0xff5 # r
norm4      = 0xff4 # r
norm2      = 0xff3 # r
acc_sop    = 0xff3 # w
norm1      = 0xff2 # r
mfb        = 0xff2 # w
prod23     = 0xff1 # r
mfa        = 0xff1 # w
prod11     = 0xff0 # r
rnd_sop    = 0xff0 # w

# pio
poi        = 0xfef # w, po invert bits
pos        = 0xfee # w, po set bits
pi         = 0xfed # r
poc        = 0xfed # w, po clear bits
po         = 0xfec # rw

# uart
# useless @ 0xfeb, yields zero if read
uart       = 0xfea # rw
tx_status  = 0xfe9 # r
rx_status  = 0xfe8 # r

# spi
spi        = 0xfe7 # rw
spi_status = 0xfe6 # r
spi_ctl    = 0xfe6 # w

# the easily-decoded base address for separating code/data from mmio
mmio       = 0xfe0 # start of mmio

asm(srcfile)
assert lptr < mmio
print('You can snoop around with dir(), dump_ram(start_adr, end_adr), globals(), etc.')
print("To save a serial bootloader app, type save_app('apps/yourapp.app')")
#print("To save an FPGA ram image (e.g. the serial loader), type save_hex('FPGAburn/ram.hex')")
print('Type exit() when done.')
