// Tiger Single Board Computer on the Lattice iCE40-HX8K breakout board

module TigerSBC (
  // control and status
  input clk_i,        // pll refence (12 MHz)
  input nrst_i,       // active-LOW, pulled up
  output [9:2] led_o, // 8 red LEDs D9..2 subscripted per the pcb silkscreen
  //
  // uart <-> host via usb, also the only convenient source of +5v
  input rxd_i,
  output txd_o,
  //
  // spi
  input miso_i, // pulled up
  output mosi_o,
  output sck_o,
  output [7:1] ncs_o,
  //
  // parallel io
  input [11:0] pi_i, // pulled up
  output [11:0] po_o
);

// led_o[9] will be driven by cpu call_overflow
// led_o[8] will be driven by cpu ret_underflow
// led_o[7] will be driven by a copy of po_o[0]
// turn off 5 unused leds
assign led_o[6:2] = 5'b0;

wire clk; // from pll
reg rst;
always @(posedge clk) rst <= ~nrst_i;

iCE40PLL_pll iCE40PLL_pll_inst(
  .RESET(1'b1), // active-LOW
  .REFERENCECLK(clk_i),
  .PLLOUTCORE(),
  .PLLOUTGLOBAL(clk),
  .LOCK()
);

// CPU memory access controls
wire loadex; // instruction packet fetches are not loads
wire storex;
wire [11:0] adr, wdat;

// adr decodes and ORable read data
//
wire ram_sel = adr[11:5] != 7'b1111111; // 000..fdf, mmio starts at fe0
wire [11:0] ram_o; // errant instruction packet fetches from mmio shall yield NOPs and wrap to the reset code
//
wire spi_sel = adr[11:1] == 11'b11111110011; // fe6..fe7
wire [7:0] _spi_o;
wire [11:0] spi_o = {4'b0, _spi_o}; // zero-extended
//
wire uart_sel = adr[11:2] == 10'b1111111010; // fe8..feb
wire [7:0] _uart_o;
wire [11:0] uart_o = {4'b0, _uart_o}; // zero-extended
//
wire pio_sel = adr[11:2] == 10'b1111111011; // fec..fef
wire [11:0] pio_o;
//
wire macc_sel = adr[11:3] == 9'b111111110; // ff0..ff7
wire [11:0] macc_o;
//
wire timer_sel = adr[11:3] == 9'b111111111; // ff8..fff
wire [11:0] timer_o;
//
// the memory read result to the cpu
wire [11:0] rdat = ram_o | spi_o | uart_o | pio_o | macc_o | timer_o;

TigerCPU cpu1 (
  .clk_i(clk),
  .rst_i(rst),
  .dat_i(rdat),
  .call_overflow_o(led_o[9]),
  .ret_underflow_o(led_o[8]),
  .storex_o(storex),
  .loadex_o(loadex),
  .adr_o(adr),
  .dat_o(wdat)
);

ram ram1 (
  .clk_i(clk),
  .sel_i(ram_sel),
  .adr_i(adr),
  .storex_i(storex),
  .dat_i(wdat),
  .dat_o(ram_o)
);

spi spi1 (
  .clk_i(clk),
  .rst_i(rst),
  .sel_i(spi_sel),
  .adr_i(adr[0]),
  .storex_i(storex),
  .dat_i(wdat[7:0]),
  .loadex_i(loadex),
  .dat_o(_spi_o),
  .miso_i(miso_i),
  .sck_o(sck_o),
  .mosi_o(mosi_o),
  .ncs_o(ncs_o)
);

uart uart1 (
  .clk_i(clk),
  .rst_i(rst),
  .sel_i(uart_sel),
  .adr_i(adr[1:0]),
  .storex_i(storex),
  .dat_i(wdat[7:0]),
  .loadex_i(loadex),
  .dat_o(_uart_o),
  .rxd_i(rxd_i),
  .txd_o(txd_o)
);

pio pio1 (
  .clk_i(clk),
  .rst_i(rst),
  .sel_i(pio_sel),
  .adr_i(adr[1:0]),
  .storex_i(storex),
  .dat_i(wdat),
  .loadex_i(loadex),
  .dat_o(pio_o),
  .pi_i(pi_i),
  .po_o(po_o),
  .led_o(led_o[7])
);

macc macc1 (
  .clk_i(clk),
  .sel_i(macc_sel),
  .adr_i(adr[2:0]),
  .storex_i(storex),
  .dat_i(wdat),
  .loadex_i(loadex),
  .dat_o(macc_o)
);

timer timer1 (
  .clk_i(clk),
  .rst_i(rst),
  .sel_i(timer_sel),
  .adr_i(adr[2:0]),
  .loadex_i(loadex),
  .dat_o(timer_o)
);

endmodule
