// the Tiger programming model is
//
//   a data stack (d:) of 16 12-bit registers, d[15] .. d[0]
//     the top of the data stack is d[0]
//     the entire context may be abandoned (being registers) at the programmer's discretion
//     there is no CPU hardware protection for data stack overflow/underflow
//
//   a program counter stack (p:) of 8 12-bit registers, p[7] .. p[0]
//     the top of the program counter stack (p[0]) is the program counter
//     the program counter stack is protected by CPU hardware
//     up to 7 calls may be nested, any deeper will cause a halt with the call_overflow LED lit
//     an unbalanced ret will cause a halt with the ret_underflow LED lit
//
//   a physical memory space of 4096 12-bit code/data/mmio locations
//     mmio may (is expected to) provide "move machine" accelerators
//
// there is no program status word, but see ADX re arithmetic carry
// all instructions execute in 1 clock cycle, including JZ whether the jump is taken or not
//
// code locations contain a packet of three 4-bit instructions
// on each instruction packet fetch, the program counter is incremented to point to the next instruction packet,
// or the first of any inline data that satisfies instances of the LIT instruction in the current packet
//
// the behaviour of each instruction is described using the following Forth-style notation
//   ->  denotes the arguments/results separator
//   d   denotes data
//   adr denotes a pointer to code/data/mmio
//   pc  denotes a program counter value
//   d:  denotes that the data stack is specified
//   p:  denotes that the program counter stack is specified

localparam NOP  = 4'h0;
// no operation
// this is necessary for the instruction packet sequencing/prefetch logic, see TigerCPU.v

localparam LIT  = 4'h1;
// d: -> d
// p: pc -> pc+1
// push d addressed by pc, increment pc
// this is the only way to convey a 12-bit assembly-time expression to a run-time constant

localparam LD   = 4'h2;
// d: adr -> d
// load d from adr

localparam ST   = 4'h3;
// d: d adr ->
// store d to adr

localparam TT   = 4'h4;
// d: d -> d d
// duplicate d

localparam INV  = 4'h5;
// d: d -> ~d
// invert d

localparam INC  = 4'h6;
// d: d -> d+1
// increment d

localparam ASR  = 4'h7;
// d: d -> d>>1
// arithmetic shift right d, the 2's-complement sign is preserved

localparam ADX  = 4'h8;
// d: d1 d2 -> d1+d2 carry
// add extended, carry is 0 or 1 and may be interpreted as not borrow

localparam ADD  = 4'h9;
// d: d1 d2 -> d1+d2
// add

localparam BAND = 4'hA;
// d: d1 d2 -> d1&d2
// binary and
// this is useful for testing which bits are one, or for zeroing bits

localparam BXOR = 4'hB;
// d: d1 d2 -> d1^d2
// binary exclusive-or
// this is useful for testing which bits are different, or for inverting bits

localparam JZ   = 4'hC;
// d: flag adr ->
// p: pc -> flag?pc:adr+1
// if false flag jump to adr, else continue any trailing instruction slots
// this is the only conditional program flow control instruction

localparam JMP  = 4'hD;
// d: adr ->
// p: pc -> adr+1
// unconditional jump to adr, abandon any trailing instruction slots

localparam CALL = 4'hE;
// d: adr ->
// p: pc -> pc adr+1
// unconditional call adr, abandon any trailing instruction slots
// halt if the program counter stack would have overflowed, lighting the dedicated red LED

localparam RET  = 4'hF;
// p: adr pc -> adr+1
// unconditional return from subroutine, abandon any trailing instruction slots
// halt if the program counter stack would have underflowed, lighting the dedicated red LED
