/*
7-channel 8-bit mode-0 master-SPI interface
note that ctl_we, tx_we, and done are internal signals for reference only
         _                                                                                  _
ctl_we _/ \________________________________________________________________________________/ \__
             _                                       _
tx_we  _____/ \_____________________________________/ \_________________________________________
       _______                                 _______                                 _________
done          \_______________________________/       \_______________________________/
       ___                                                                                    __
ncsx      \__________________________________________________________________________________/
                 _   _   _   _   _   _   _   _           _   _   _   _   _   _   _   _
sck    _________/7\_/6\_/5\_/4\_/3\_/2\_/1\_/0\_________/7\_/6\_/5\_/4\_/3\_/2\_/1\_/0\_________
       _______ ___ ___ ___ ___ ___ ___ ___ ___________ ___ ___ ___ ___ ___ ___ ___ _____________
mosi   _______X___X___X___X___X___X___X___X___________X___X___X___X___X___X___X___X_____________
           _______ ___ ___ ___ ___ ___ ___ ___ ___________ ___ ___ ___ ___ ___ ___ ___ ______
miso   ---<_______X___X___X___X___X___X___X___X___________X___X___X___X___X___X___X___X______>--
*/

module spi (
  // CPU interface
  input clk_i,
  input rst_i,
  input sel_i,       // 2-aligned
  input adr_i,       // subselect
  input storex_i,
  input [7:0] dat_i,
  input loadex_i,
  output [7:0] dat_o,
  // SPI interface
  input miso_i,      // master in slave out
  output sck_o,      // serial clock
  output mosi_o,     // master out slave in
  output [7:1] ncs_o // active-LOW chip selects
);

wire ctl_we = sel_i & storex_i & ~adr_i; // spi_ctl
wire tx_we  = sel_i & storex_i & adr_i; // spi

// control
reg [7:0] ctl;
always @(posedge clk_i)
  if (rst_i) ctl <= 8'b11111000; // slowest sck, no device selected
  else if (ctl_we) ctl <= dat_i;
wire [4:0] speed = ctl[7:3]; // fsck = fclk/(2*(speed+1))
wire [2:0] cs    = ctl[2:0]; // SPI device selector, 0 is none
assign ncs_o[7] = ~(cs == 7);
assign ncs_o[6] = ~(cs == 6);
assign ncs_o[5] = ~(cs == 5);
assign ncs_o[4] = ~(cs == 4);
assign ncs_o[3] = ~(cs == 3);
assign ncs_o[2] = ~(cs == 2);
assign ncs_o[1] = ~(cs == 1);

// sck edge bit rate generator
reg [4:0] brg;
wire tc = brg == speed; // terminal count
always @(posedge clk_i)
  if (tx_we | tc) brg <= 0;
  else brg <= brg + 1;

// sck edge counter
reg [4:0] sckedge; // 0..16
wire done = sckedge[4]; // sckedge reached 16
wire sck = sckedge[0]; // the current mode 0 sck phase
assign sck_o = sck;
always @(posedge clk_i)
  if (rst_i) sckedge <= 16; // transaction done
  else if (tx_we) sckedge <= 0; // start transaction
  else if (~done & tc) sckedge <= sckedge + 1;

// master out slave in
reg [7:0] mosi;
assign mosi_o = mosi[7];
always @(posedge clk_i)
  if (tx_we) mosi <= dat_i;
  else if (~done & sck & tc) mosi <= {mosi[6:0], mosi[0]}; // transmit on sck falling edge

// master in slave out
reg [7:0] miso;
always @(posedge clk_i) if (~done & ~sck & tc) miso <= {miso[6:0], miso_i}; // receive on sck rising edge

// read operations
reg [7:0] dat_o;
always @*
  if (sel_i & loadex_i)
    case (adr_i)
      0: dat_o = {7'b0, done}; // spi_status
      1: dat_o = miso; // spi
    endcase
  else dat_o = 0; // ORable

endmodule
